import typing

import zope.interface
from attrs import resolve_types, validators, field, NOTHING, setters


def _union_type_validator(types):
    validators = tuple(_type_validator(t) for t in types)

    def check(self, attribute, value):
        for validate in validators:
            try:
                validate(self, attribute, value)
            except TypeError as e:
                pass
            else:
                return
        raise TypeError(
            "%s (%s) is not one of types: %s"
            % (value, type(value), ", ".join("%s" % t for t in types))
        )

    return check


def _type_validator(t):
    original_type = getattr(t, "__origin__", None)
    if type(t) is type:
        return validators.instance_of(t)
    elif type(t) is zope.interface.interface.InterfaceClass:
        return validators.provides(t)
    elif original_type is list or original_type is typing.List:
        return validators.deep_iterable(member_validator=_type_validator(t.__args__[0]))
    elif original_type is dict or original_type is typing.Dict:
        return validators.deep_mapping(
            key_validator=_type_validator(t.__args__[0]),
            value_validator=_type_validator(t.__args__[1]),
        )
    elif original_type is typing.Union:
        return _union_type_validator(t.__args__)
    raise TypeError("unanticipated type %s (%s)" % (t, getattr(t, "__origin__")))


def field_transformer(namespace):
    def transform(cls, fields):
        results = []
        resolve_types(
            cls, globalns=namespace, localns={cls.__name__: cls}, attribs=fields
        )
        for field in fields:
            if type(field.type) is str and field.type == cls.__name__:
                field = field.evolve(type=cls)
            validator = _type_validator(field.type)
            validator = (
                validator
                if field.validator is None
                else validators.and_(validator, field.validator)
            )
            if field.default is NOTHING and not field.metadata.get(REQUIRED_K, False):
                validator = validators.optional(validator)
                field = field.evolve(default=None)
            on_setattr = setters.validate
            on_setattr = (
                on_setattr
                if field.on_setattr is None
                else setters.pipe(on_setattr, field.on_setattr)
            )
            results.append(field.evolve(validator=validator, on_setattr=on_setattr))
        return results

    return transform


DOC_K = "deba_doc"
REQUIRED_K = "deba_required"


def doc(
    doc: str,
    default=NOTHING,
    factory=None,
    converter=None,
    required=False,
    validator=None,
):
    return field(
        metadata={DOC_K: doc, REQUIRED_K: required},
        default=default,
        factory=factory,
        converter=converter,
        validator=validator,
    )



import re
import itertools
import typing
import ast
from fnmatch import fnmatchcase

from attrs import define, field
from deba.attrs_utils import field_transformer, doc
from deba.deps.module import Stack


class ExprTemplateParseError(ValueError):
    pass


@define(field_transformer=field_transformer(globals()))
class ExprPattern(object):
    node: ast.AST = field()
    text: str = field()
    file_pat: object = field()
    patterns: typing.List[str] = field(factory=list)
    backtick_name_pat: typing.ClassVar[object] = re.compile(
        r"^deba_backtick_pat_(\d{3})$"
    )

    @node.validator
    def is_expr_valid(self, attr, value: ast.AST):
        str_count = 0
        if not isinstance(value, ast.Call):
            raise ExprTemplateParseError(
                "expression must be a function call, found %s" % type(value)
            )
        for node in ast.walk(value):
            if isinstance(node, ast.Constant):
                if type(node.value) is not str:
                    raise ExprTemplateParseError(
                        "expect exactly 1 string constant, found %s" % type(node.value)
                    )
                str_count += 1
                try:
                    re.compile(node.value)
                except re.error as e:
                    raise ExprTemplateParseError(
                        "invalid regular expression r'%s': %s" % (node.value, e.args[0])
                    )
                s = node.value if node.value[0] == "^" else "^" + node.value
                s = s if s[-1] == "$" else s + "$"
                self.file_pat = re.compile(s)
            elif isinstance(node, ast.Call):
                args = getattr(node, "args", [])
                keywords = getattr(node, "keywords", [])
                if (
                    (len(args) == 0 and len(keywords) == 0)
                    or len(args) > 1
                    or len(keywords) > 1
                ):
                    raise ExprTemplateParseError(
                        "function call must have exactly one argument or one keyword argument"
                    )
        if str_count != 1:
            raise ExprTemplateParseError(
                "expect exactly 1 string, found %d" % str_count
            )

    @classmethod
    def from_str(cls, text: str) -> "ExprPattern":
        patterns = []
        orig_text = text
        while True:
            try:
                mod = ast.parse(text)
            except SyntaxError as e:
                lines = text.split("\n")
                line = lines[e.lineno - 1]
                c = line[e.offset - 1]
                if c == "`":
                    next_backtick = line[e.offset :].find("`")
                    if next_backtick == -1:
                        raise ExprTemplateParseError(
                            "backtick pattern not closed at line %d, offset %d"
                            % (e.lineno, e.offset)
                        )
                    next_backtick += e.offset
                    pattern = line[e.offset : next_backtick]
                    lines[e.lineno - 1] = "%sdeba_backtick_pat_%03d%s" % (
                        line[: e.offset - 1],
                        len(patterns),
                        line[next_backtick + 1 :],
                    )
                    patterns.append(pattern)
                    text = "\n".join(lines)
                else:
                    raise
            else:
                if len(mod.body) > 1:
                    raise ExprTemplateParseError(
                        "expect exactly 1 expression, found %d" % len(mod.body)
                    )
                return cls(text=orig_text, patterns=patterns, node=mod.body[0].value)

    def match_constant(
        self, node1: ast.Constant, node2: ast.Constant
    ) -> typing.Tuple[str, bool]:
        if type(node1.value) is str and type(node2.value) is str:
            m = self.file_pat.match(node2.value)
            if m is None:
                return "", False
            return node2.value, True
        return "", node1.value == node2.value

    def match_ast(
        self, scopes: Stack, node1: ast.AST, node2: ast.AST
    ) -> typing.Tuple[str, bool]:
        if isinstance(node1, ast.Constant):
            if isinstance(node2, ast.Constant):
                return self.match_constant(node1, node2)
            node = scopes.dereference(node2)
            if node is None or not isinstance(node.ast, ast.Constant):
                return "", False
            return self.match_constant(node1, node.ast)

        if type(node1) is not type(node2):
            return "", False

        if isinstance(node1, ast.AST):
            if isinstance(node1, ast.Name):
                if type(node1.ctx) is not type(node2.ctx):
                    return "", False
                if node1.id == node2.id:
                    return "", True
                m = ExprPattern.backtick_name_pat.match(node1.id)
                if m is None:
                    return "", False
                pat = self.patterns[int(m.group(1))]
                return "", fnmatchcase(node2.id, pat)
            elif isinstance(node1, ast.Call):
                _, ok = self.match_ast(scopes, node1.func, node2.func)
                if not ok:
                    return "", False
                s = ""
                if len(node1.args) > 0:
                    for arg in node2.args:
                        v, ok = self.match_ast(scopes, node1.args[0], arg)
                        if ok:
                            if v != "":
                                s = v
                            break
                    else:
                        return "", False
                elif len(node1.keywords) > 0:
                    for kw in node2.keywords:
                        v, ok = self.match_ast(scopes, node1.keywords[0], kw)
                        if ok:
                            if v != "":
                                s = v
                            break
                    else:
                        return "", False
                return s, True
            s = ""
            for k, v in ast.iter_fields(node1):
                v, ok = self.match_ast(scopes, v, getattr(node2, k))
                if not ok:
                    return "", False
                elif v != "":
                    s = v
            return s, True
        elif isinstance(node1, list):
            s = ""
            for v, ok in itertools.starmap(
                lambda x, y: self.match_ast(scopes, x, y),
                itertools.zip_longest(node1, node2),
            ):
                if not ok:
                    return "", False
                elif v != "":
                    s = v
            return s, True
        else:
            return "", node1 == node2

    def match_node(self, scopes: Stack, node: ast.AST) -> typing.Union[None, str]:
        s, ok = self.match_ast(scopes, self.node, node)
        if ok:
            return s


def expr_templates(l: typing.Union[typing.List[str], None]) -> typing.List[ExprPattern]:
    return (
        None
        if l is None
        else [s if isinstance(s, ExprPattern) else ExprPattern.from_str(s) for s in l]
    )


@define(field_transformer=field_transformer(globals()))
class ExprPatterns(object):
    """Expression patterns."""

    prerequisites: typing.List[ExprPattern] = doc(
        "prerequisite expression patterns", converter=expr_templates
    )
    targets: typing.List[ExprPattern] = doc(
        "target expression patterns", converter=expr_templates
    )

    def as_dict(self) -> typing.Dict:
        return {
            "prerequisites": [obj.text for obj in self.prerequisites],
            "targets": [obj.text for obj in self.targets],
        }



from unittest import TestCase
import typing

from attrs import define, field

import zope.interface
from deba.attrs_utils import field_transformer


class IC(zope.interface.Interface):
    def get_c():
        """get c"""


@define(field_transformer=field_transformer(globals()))
class A:
    a: str
    b: int
    c: typing.Union[str, typing.List[str]]
    d: float = field(default=3.14)


@zope.interface.implementer(IC)
@define(field_transformer=field_transformer(globals()))
class C:
    c: int

    def get_c(self):
        return self.c


@define(field_transformer=field_transformer(globals()))
class B:
    a: A
    b: typing.List[A]
    c: typing.Dict[str, A]
    d: IC
    e: typing.List[IC]
    f: typing.Dict[str, IC]


class FieldTransformerTestCase(TestCase):
    def test_scalar_fields(self):
        obj = A()
        self.assertIsNone(obj.a)
        self.assertIsNone(obj.b)
        self.assertEqual(obj.d, 3.14)

        obj = A("a", 1, d=12.3)
        self.assertEqual(obj.a, "a")
        self.assertEqual(obj.b, 1)
        self.assertEqual(obj.d, 12.3)

        with self.assertRaises(TypeError):
            obj = A(a=3)
        with self.assertRaises(TypeError):
            obj = A(b="3")

    def test_union_fields(self):
        obj = A(c="3")
        self.assertEqual(obj.c, "3")

        obj = A(c=["3"])
        self.assertEqual(obj.c, ["3"])

        with self.assertRaises(TypeError):
            obj = A(c=3.1)

    def test_nested_fields(self):
        obj = B()
        self.assertIsNone(obj.a)
        self.assertIsNone(obj.b)
        self.assertIsNone(obj.c)

        obj = B(a=A(), b=[], c=dict())
        self.assertEqual(obj.a, A())
        self.assertEqual(obj.b, [])
        self.assertEqual(obj.c, dict())

        obj = B(b=[A()], c={"a": A()})
        self.assertEqual(obj.b, [A()])
        self.assertEqual(obj.c, {"a": A()})

        with self.assertRaises(TypeError):
            obj = B(a=B())
        with self.assertRaises(TypeError):
            obj = B(b=A())
        with self.assertRaises(TypeError):
            obj = B(b=[B()])
        with self.assertRaises(TypeError):
            obj = B(c=A())
        with self.assertRaises(TypeError):
            obj = B(c={"a": B()})

    def test_interface_fields(self):
        obj = B(d=C())
        self.assertEqual(obj.d, C())
        with self.assertRaises(TypeError):
            obj = B(d=A())

        obj = B(e=[C()])
        self.assertEqual(obj.e, [C()])
        with self.assertRaises(TypeError):
            obj = B(e=[A()])

        obj = B(f={"a": C()})
        self.assertEqual(obj.f, {"a": C()})
        with self.assertRaises(TypeError):
            obj = B(f={"a": A()})



import re
import typing

import attr
import yaml
from yaml.dumper import Dumper


def to_camel_case(s):
    components = s.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


word_break_pattern = re.compile(r"([a-z])([A-Z])")


def to_snake_case(s):
    return word_break_pattern.sub(r"\1_\2", s).lower()


def noop(self, *args, **kw):
    pass


yaml.emitter.Emitter.process_tag = noop


def _object_as_dict(obj):
    if attr.has(obj.__class__):
        if hasattr(obj, "as_dict"):
            return obj.as_dict()
        fields = attr.fields(obj.__class__)
        return {
            to_camel_case(field.name): _object_as_dict(getattr(obj, field.name))
            for field in fields
            if not field.name.startswith("_")
            and getattr(obj, field.name, None) is not None
        }
    return obj


def represent_attr_object(dumper, data):
    if attr.has(data.__class__):
        return dumper.represent_dict(_object_as_dict(data))
    return dumper.represent_object(data)


yaml.add_multi_representer(object, represent_attr_object)


def yaml_dump(data, **kwargs) -> str:
    return yaml.dump(data, Dumper=Dumper, **kwargs)


def yaml_load(s, serializer_cls):
    data = yaml.load(s, Loader=yaml.Loader)
    return _deserialize(data, serializer_cls)


def _deserialize(data, serializer_cls):
    kwargs = dict()
    fields_dict = attr.fields_dict(serializer_cls)
    if type(data) is str:
        return serializer_cls.from_str(data)
    data = [(to_snake_case(k), v) for k, v in data.items()]
    data_stack = [(kwargs, k, fields_dict[k], v) for k, v in data if k != "meta"]
    while len(data_stack) > 0:
        parent, name, field, value = data_stack.pop()
        if value is None:
            parent[name] = value
        elif type(field.type) is type:
            if attr.has(field.type):
                parent[name] = _deserialize(value, field.type)
            else:
                parent[name] = value
        elif field.type.__origin__ is list or field.type.__origin__ is typing.List:
            el_cls = field.type.__args__[0]
            if type(value) is not list:
                raise TypeError(
                    'keyword argument "%s" should be a list of %s' % (name, el_cls)
                )
            if attr.has(el_cls):
                parent[name] = [_deserialize(e, el_cls) for e in value]
            else:
                parent[name] = value
        elif field.type.__origin__ is dict or field.type.__origin__ is typing.Dict:
            el_cls = field.type.__args__[1]
            if type(value) is not dict:
                raise TypeError(
                    'keyword argument "%s" should be a dict of %s' % (name, el_cls)
                )
            if attr.has(el_cls):
                parent[name] = {k: _deserialize(e, el_cls) for k, e in value.items()}
            else:
                parent[name] = value
        else:
            raise TypeError("unanticipated field type %s" % field.type)
    return serializer_cls(**kwargs)



import ast
import os
from importlib.machinery import ModuleSpec
import typing

from deba.deps.module import Loader, Node, Stack

from deba.deps.expr import ExprPattern


class ModuleParseError(ValueError):
    pass


def is_name(node, id) -> bool:
    return isinstance(node, ast.Name) and node.id == id


def is_const(node, val) -> bool:
    return isinstance(node, ast.Constant) and node.value == val


def is_main_block(stmt: ast.stmt) -> bool:
    return (
        isinstance(stmt, ast.If)
        and is_name(stmt.test.left, "__name__")
        and isinstance(stmt.test.ops[0], ast.Eq)
        and is_const(stmt.test.comparators[0], "__main__")
    )


def trim_suffix(s: str, suffix: str) -> str:
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def build_module_from_filepath(loader: Loader, filepath: str) -> Node:
    if not os.path.isfile(filepath):
        if not os.path.isdir(filepath) or not os.path.isfile(
            os.path.join(filepath, "__init__.py")
        ):
            raise ValueError("not a file path: %s" % filepath)
    parent, filename = os.path.split(filepath)
    node = loader.find_module(trim_suffix(filename, ".py"), [parent])
    return node


def find_dependencies(
    loader: Loader,
    filepath: str,
    prerequisite_targets: typing.List[ExprPattern],
    target_patterns: typing.List[ExprPattern],
) -> typing.Tuple[typing.List[str], typing.List[str]]:
    module_node = build_module_from_filepath(loader, filepath)
    stack = Stack()
    for stmt in module_node.ast.body:
        if is_main_block(stmt):
            return scan(
                loader,
                stmt,
                module_node.spec,
                stack.push(),
                prerequisite_targets,
                target_patterns,
            )
        else:
            loader.populate_scope(module_node.spec, stack, module_node.ast, stmt)
    raise ModuleParseError("main block not found")


def scan(
    loader: Loader,
    node: ast.AST,
    spec: ModuleSpec,
    stack: Stack,
    prerequisite_patterns: typing.List[ExprPattern],
    target_patterns: typing.List[ExprPattern],
) -> typing.Tuple[typing.List[str], typing.List[str]]:
    prerequisites, targets = [], []
    for stmt in node.body:
        loader.populate_scope(spec, stack, node, stmt)
        for t in ast.walk(stmt):
            if isinstance(t, ast.Call):
                for tmpl in target_patterns:
                    s = tmpl.match_node(stack, t)
                    if s is not None:
                        targets.append(s)
                        break
                else:
                    for tmpl in prerequisite_patterns:
                        s = tmpl.match_node(stack, t)
                        if s is not None:
                            prerequisites.append(s)
                            break
                    else:
                        func = stack.dereference(t.func)
                        if func is None or not isinstance(func.ast, ast.FunctionDef):
                            continue
                        ins, outs = scan(
                            loader,
                            func.ast,
                            func.spec,
                            stack.push(),
                            prerequisite_patterns,
                            target_patterns,
                        )
                        prerequisites = prerequisites + ins
                        targets = targets + outs
    return prerequisites, targets



import typing
import tempfile
import ast
import io
import os
import sys
import shutil

from attrs import define, field


def ast_dump(obj):
    if sys.version_info[1] < 9:
        ast.dump(obj)
    else:
        ast.dump(obj, indent=4)


class ASTMixin(object):
    def assertASTEqual(
        self, a: ast.AST, b: ast.AST, msg: typing.Union[str, None] = None
    ):
        if a is None or b is None:
            self.assertEqual(a, b, msg)
        else:
            self.assertEqual(ast_dump(a), ast_dump(b), msg)


class TempDirMixin(object):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._dir = tempfile.TemporaryDirectory()

    def setUp(self):
        super().setUp()
        self.maxDiff = None
        for filename in os.listdir(self._dir.name):
            file_path = os.path.join(self._dir.name, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

    @classmethod
    def tearDownClass(cls) -> None:
        cls._dir.cleanup()
        return super().tearDownClass()

    def file_path(self, filename: str) -> str:
        return os.path.join(self._dir.name, filename)

    def write_file(self, filename: str, lines: typing.List[str]):
        os.makedirs(os.path.dirname(self.file_path(filename)), exist_ok=True)
        with open(self.file_path(filename), "w") as f:
            f.write("\n".join(lines))

    def mod_time(self, filename: str) -> float:
        return os.path.getmtime(self.file_path(filename))

    def assertFileContent(self, filename: str, lines: typing.List[str]) -> float:
        """Asserts file content and returns modified time as seconds since the epoch"""
        with open(self.file_path(filename), "r") as f:
            self.assertEqual(
                f.read(),
                "\n".join(lines),
            )
        return self.mod_time(filename)

    def assertFileModifiedSince(self, filename: str, time: float):
        self.assertGreater(self.mod_time(filename), time)

    def assertFileNotModifiedSince(self, filename: str, time: float):
        self.assertEqual(self.mod_time(filename), time)

    def assertFileRemoved(self, filename: str):
        self.assertFalse(os.path.isfile(self.file_path(filename)))

    def assertDirRemoved(self, dirname: str):
        self.assertFalse(os.path.isdir(self.file_path(dirname)))



from importlib.machinery import PathFinder
import typing
from unittest import TestCase
import os
import ast

import prettyprinter

from deba.deps.expr import ExprPattern
from deba.deps.module import Node, Module, Package, Loader, trim_suffix
from deba.deps.find import build_module_from_filepath, find_dependencies
from deba.test_utils import ASTMixin, TempDirMixin


prettyprinter.install_extras(["attrs"])


class DepsFinderTestCase(ASTMixin, TempDirMixin, TestCase):
    def assertDictEqual(
        self,
        a: typing.Union[typing.Dict, None],
        b: typing.Union[typing.Dict, None],
        msg: typing.Union[str, None] = None,
    ):
        if a is None:
            self.assertIsNone(b, msg)
        else:
            for k in b:
                if k not in a:
                    raise AssertionError("%s: %s not found in %s" % (k, b[k], a))
            for k, v in a.items():
                if k not in b:
                    raise AssertionError("%s: %s not found in %s" % (k, v, b))
                self.assertObjectEqual(v, b[k], msg)

    def assertObjectEqual(
        self,
        a: typing.Union[Node, None],
        b: typing.Union[Node, None],
        msg: typing.Union[str, None] = None,
    ):
        try:
            self.assertEqual(type(a), type(b))
            if isinstance(a, Package):
                self.assertDictEqual(a.modules, b.modules)
            elif isinstance(a, Module):
                self.assertASTEqual(
                    a.ast,
                    b.ast,
                )
                self.assertEqual(a.spec, b.spec, msg)
                self.assertDictEqual(a.children, b.children, msg)
            elif isinstance(b, Node):
                self.assertASTEqual(a.ast, b.ast)
                self.assertDictEqual(a.children, b.children, msg)
        except Exception:
            prettyprinter.cpprint(a)
            prettyprinter.cpprint(b)
            raise

    def test_find_spec(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["a = 1"])
        self.write_file("b/__init__.py", [""])
        self.write_file("b/c.py", ["b = r'abc'"])

        spec = loader.find_spec("a")
        self.assertIsNotNone(spec)
        self.assertEqual(spec.origin, self.file_path("a.py"))

        spec = loader.find_spec("b")
        self.assertIsNotNone(spec)
        self.assertEqual(spec.origin, self.file_path("b/__init__.py"))

        self.assertIsNone(loader.find_spec("c"))

        spec = loader.find_spec("c", spec.submodule_search_locations)
        self.assertIsNotNone(spec)
        self.assertEqual(spec.origin, self.file_path("b/c.py"))

        spec = loader.find_spec("b.c")
        self.assertIsNotNone(spec)
        self.assertEqual(spec.origin, self.file_path("b/c.py"))

        self.assertIsNone(loader.find_spec("b.d"))

    def test_parse_module_spec(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["a = 1"])
        mod = loader.parse_ast(self.file_path("a.py"))
        self.assertASTEqual(
            mod,
            ast.Module(
                [
                    ast.Assign(
                        targets=[ast.Name(id="a", ctx=ast.Store())],
                        value=ast.Constant(value=1),
                    )
                ],
                type_ignores=[],
            ),
        )

        os.remove(self.file_path("a.py"))
        # returns from cache
        self.assertASTEqual(loader.parse_ast(self.file_path("a.py")), mod)

    def spec(self, s: str):
        path, name = os.path.split(self.file_path(s))
        return PathFinder.find_spec(trim_suffix(name, ".py"), [path])

    def test_build_module_from_filepath(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["a = 1"])
        self.write_file("b/__init__.py", ["b = r'b'"])
        self.write_file("b/c.py", ["b = r'abc'"])
        self.write_file("b/d/__init__.py", ["d = r'd'"])
        self.write_file("b/d/e.py", ["e = 2"])

        node_a = build_module_from_filepath(loader, self.file_path("a.py"))
        self.assertObjectEqual(
            node_a,
            Module(
                ast.Module(
                    [
                        ast.Assign(
                            targets=[ast.Name(id="a", ctx=ast.Store())],
                            value=ast.Constant(value=1),
                        )
                    ],
                    type_ignores=[],
                ),
                spec=self.spec("a.py"),
                children={"a": Node(ast.Constant(value=1), self.spec("a.py"))},
            ),
        )

        node_b = build_module_from_filepath(loader, self.file_path("b"))
        self.assertEqual(
            node_b.modules["__init__"].spec.origin, self.file_path("b/__init__.py")
        )
        self.assertObjectEqual(
            node_b,
            Package(
                {
                    "__init__": Module(
                        ast.Module(
                            [
                                ast.Assign(
                                    targets=[ast.Name(id="b", ctx=ast.Store())],
                                    value=ast.Constant(value="b"),
                                )
                            ],
                            type_ignores=[],
                        ),
                        self.spec("b"),
                        children={
                            "b": Node(
                                ast.Constant(value="b"),
                                self.spec("b"),
                            ),
                        },
                    ),
                    "c": Module(
                        ast.Module(
                            [
                                ast.Assign(
                                    targets=[ast.Name(id="b", ctx=ast.Store())],
                                    value=ast.Constant(value="abc"),
                                )
                            ],
                            type_ignores=[],
                        ),
                        self.spec("b/c.py"),
                        children={
                            "b": Node(ast.Constant(value="abc"), self.spec("b/c.py"))
                        },
                    ),
                    "d": Package(
                        {
                            "__init__": Module(
                                ast.Module(
                                    [
                                        ast.Assign(
                                            targets=[ast.Name(id="d", ctx=ast.Store())],
                                            value=ast.Constant(value="d"),
                                        )
                                    ],
                                    type_ignores=[],
                                ),
                                self.spec("b/d"),
                                children={
                                    "d": Node(
                                        ast.Constant(value="d"),
                                        self.spec("b/d"),
                                    ),
                                },
                            ),
                            "e": Module(
                                ast.Module(
                                    [
                                        ast.Assign(
                                            targets=[ast.Name(id="e", ctx=ast.Store())],
                                            value=ast.Constant(value=2),
                                        )
                                    ],
                                    type_ignores=[],
                                ),
                                self.spec("b/d/e.py"),
                                children={
                                    "e": Node(
                                        ast.Constant(value=2),
                                        self.spec("b/d/e.py"),
                                    )
                                },
                            ),
                        }
                    ),
                }
            ),
        )

    def test_import_package(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["my_a = 1"])
        self.write_file("b.py", ["import a"])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("b.py")),
            Module(
                ast.Module(
                    body=[ast.Import(names=[ast.alias(name="a")])],
                    type_ignores=[],
                ),
                spec=self.spec("b"),
                children={
                    "a": Module(
                        ast.Module(
                            [
                                ast.Assign(
                                    targets=[ast.Name(id="my_a", ctx=ast.Store())],
                                    value=ast.Constant(value=1),
                                )
                            ],
                            type_ignores=[],
                        ),
                        spec=self.spec("a"),
                        children={"my_a": Node(ast.Constant(value=1), self.spec("a"))},
                    )
                },
            ),
        )

    def test_import_from(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["my_a = 1"])
        self.write_file("b.py", ["from a import my_a"])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("b.py")),
            Module(
                ast.Module(
                    body=[
                        ast.ImportFrom(
                            module="a", names=[ast.alias(name="my_a")], level=0
                        ),
                    ],
                    type_ignores=[],
                ),
                self.spec("b"),
                children={"my_a": Node(ast.Constant(value=1), self.spec("b"))},
            ),
        )

    def test_import_as(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["my_a = 1"])
        self.write_file("b.py", ["import a as a_dash"])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("b.py")),
            Module(
                ast.Module(
                    body=[ast.Import(names=[ast.alias(name="a", asname="a_dash")])],
                    type_ignores=[],
                ),
                self.spec("b"),
                children={
                    "a_dash": Module(
                        ast.Module(
                            [
                                ast.Assign(
                                    targets=[ast.Name(id="my_a", ctx=ast.Store())],
                                    value=ast.Constant(value=1),
                                )
                            ],
                            type_ignores=[],
                        ),
                        self.spec("a"),
                        children={"my_a": Node(ast.Constant(value=1), self.spec("a"))},
                    )
                },
            ),
        )

    def test_import_from_as(self):
        loader = Loader([self._dir.name])
        self.write_file("a.py", ["my_a = 1"])
        self.write_file("b.py", ["from a import my_a as my_b"])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("b.py")),
            Module(
                ast.Module(
                    body=[
                        ast.ImportFrom(
                            module="a",
                            names=[ast.alias(name="my_a", asname="my_b")],
                            level=0,
                        ),
                    ],
                    type_ignores=[],
                ),
                self.spec("b"),
                children={"my_b": Node(ast.Constant(value=1), self.spec("b"))},
            ),
        )

    def test_func_def(self):
        self.write_file(
            "a.py",
            [
                "def my_func():",
                "  pass",
                "",
            ],
        )
        loader = Loader([self._dir.name])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("a.py")),
            Module(
                ast.Module(
                    body=[
                        ast.FunctionDef(
                            name="my_func",
                            args=ast.arguments(
                                posonlyargs=[],
                                args=[],
                                kwonlyargs=[],
                                kw_defaults=[],
                                defaults=[],
                            ),
                            body=[ast.Pass()],
                            decorator_list=[],
                        ),
                    ],
                    type_ignores=[],
                ),
                self.spec("a"),
                children={
                    "my_func": Node(
                        ast.FunctionDef(
                            name="my_func",
                            args=ast.arguments(
                                posonlyargs=[],
                                args=[],
                                kwonlyargs=[],
                                kw_defaults=[],
                                defaults=[],
                            ),
                            body=[ast.Pass()],
                            decorator_list=[],
                        ),
                        self.spec("a"),
                    )
                },
            ),
        )

    def test_class_def(self):
        self.write_file(
            "a.py",
            [
                "class MyClass:",
                "  a = 'b'",
                "  c = a",
                "  d = my_b",
                "",
                "  def f(self):",
                "    pass",
                "",
                "  @classmethod",
                "  def g(cls):",
                "    pass",
                "",
                "  @staticmethod",
                "  def e():",
                "    pass",
                "",
            ],
        )
        loader = Loader([self._dir.name])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("a.py")),
            Module(
                ast.Module(
                    body=[
                        ast.ClassDef(
                            name="MyClass",
                            bases=[],
                            keywords=[],
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="a", ctx=ast.Store())],
                                    value=ast.Constant(value="b"),
                                ),
                                ast.Assign(
                                    targets=[ast.Name(id="c", ctx=ast.Store())],
                                    value=ast.Name(id="a", ctx=ast.Load()),
                                ),
                                ast.Assign(
                                    targets=[ast.Name(id="d", ctx=ast.Store())],
                                    value=ast.Name(id="my_b", ctx=ast.Load()),
                                ),
                                ast.FunctionDef(
                                    name="f",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[ast.arg(arg="self")],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[],
                                ),
                                ast.FunctionDef(
                                    name="g",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[ast.arg(arg="cls")],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="classmethod", ctx=ast.Load())
                                    ],
                                ),
                                ast.FunctionDef(
                                    name="e",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="staticmethod", ctx=ast.Load())
                                    ],
                                ),
                            ],
                            decorator_list=[],
                        ),
                    ],
                    type_ignores=[],
                ),
                self.spec("a"),
                loaded=True,
                children={
                    "MyClass": Node(
                        ast.ClassDef(
                            name="MyClass",
                            bases=[],
                            keywords=[],
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="a", ctx=ast.Store())],
                                    value=ast.Constant(value="b"),
                                ),
                                ast.Assign(
                                    targets=[ast.Name(id="c", ctx=ast.Store())],
                                    value=ast.Name(id="a", ctx=ast.Load()),
                                ),
                                ast.Assign(
                                    targets=[ast.Name(id="d", ctx=ast.Store())],
                                    value=ast.Name(id="my_b", ctx=ast.Load()),
                                ),
                                ast.FunctionDef(
                                    name="f",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[ast.arg(arg="self")],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[],
                                ),
                                ast.FunctionDef(
                                    name="g",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[ast.arg(arg="cls")],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="classmethod", ctx=ast.Load())
                                    ],
                                ),
                                ast.FunctionDef(
                                    name="e",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="staticmethod", ctx=ast.Load())
                                    ],
                                ),
                            ],
                            decorator_list=[],
                        ),
                        self.spec("a"),
                        children={
                            "a": Node(ast.Constant(value="b"), self.spec("a")),
                            "c": Node(ast.Constant(value="b"), self.spec("a")),
                            "g": Node(
                                ast.FunctionDef(
                                    name="g",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[ast.arg(arg="cls")],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="classmethod", ctx=ast.Load())
                                    ],
                                ),
                                self.spec("a"),
                            ),
                            "e": Node(
                                ast.FunctionDef(
                                    name="e",
                                    args=ast.arguments(
                                        posonlyargs=[],
                                        args=[],
                                        kwonlyargs=[],
                                        kw_defaults=[],
                                        defaults=[],
                                    ),
                                    body=[ast.Pass()],
                                    decorator_list=[
                                        ast.Name(id="staticmethod", ctx=ast.Load())
                                    ],
                                ),
                                self.spec("a"),
                            ),
                        },
                    )
                },
            ),
        )

    def test_assignment(self):
        self.write_file(
            "a.py",
            [
                "class MyClass:",
                "  a = 'b'",
                "",
                "var1 = MyClass.a",
                "var2 = var1",
                "e, f = 123, 'def'",
            ],
        )
        loader = Loader([self._dir.name])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("a.py")),
            Module(
                ast.Module(
                    body=[
                        ast.ClassDef(
                            name="MyClass",
                            bases=[],
                            keywords=[],
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="a", ctx=ast.Store())],
                                    value=ast.Constant(value="b"),
                                )
                            ],
                            decorator_list=[],
                        ),
                        ast.Assign(
                            targets=[ast.Name(id="var1", ctx=ast.Store())],
                            value=ast.Attribute(
                                value=ast.Name(id="MyClass", ctx=ast.Load()),
                                attr="a",
                                ctx=ast.Load(),
                            ),
                        ),
                        ast.Assign(
                            targets=[ast.Name(id="var2", ctx=ast.Store())],
                            value=ast.Name(id="var1", ctx=ast.Load()),
                        ),
                        ast.Assign(
                            targets=[
                                ast.Tuple(
                                    elts=[
                                        ast.Name(id="e", ctx=ast.Store()),
                                        ast.Name(id="f", ctx=ast.Store()),
                                    ],
                                    ctx=ast.Store(),
                                )
                            ],
                            value=ast.Tuple(
                                elts=[
                                    ast.Constant(value=123),
                                    ast.Constant(value="def"),
                                ],
                                ctx=ast.Load(),
                            ),
                        ),
                    ],
                    type_ignores=[],
                ),
                self.spec("a"),
                children={
                    "MyClass": Node(
                        ast.ClassDef(
                            name="MyClass",
                            bases=[],
                            keywords=[],
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="a", ctx=ast.Store())],
                                    value=ast.Constant(value="b"),
                                )
                            ],
                            decorator_list=[],
                        ),
                        self.spec("a"),
                        children={"a": Node(ast.Constant(value="b"), self.spec("a"))},
                    ),
                    "var1": Node(ast.Constant(value="b"), self.spec("a")),
                    "var2": Node(ast.Constant(value="b"), self.spec("a")),
                    "e": Node(ast.Constant(value=123), self.spec("a")),
                    "f": Node(ast.Constant(value="def"), self.spec("a")),
                },
            ),
        )

    def test_find_dependencies(self):
        loader = Loader([self._dir.name])
        self.write_file(
            "b.py",
            [
                "class MyClass:",
                "  @classmethod",
                "  def save_file(cls, a):",
                "    a.to_csv('asd.csv')",
                "",
            ],
        )
        self.write_file(
            "a.py",
            [
                "import b",
                "",
                "b_name = 'file_b.csv'",
                "",
                "def my_func():",
                "  return read_csv('qwe.csv')",
                "",
                "if __name__ == '__main__':",
                "  a = read_csv('abc.csv')",
                "  b = read_csv(b_name)",
                "  d = my_func()",
                "",
                "  a.to_csv('def.csv')",
                "  c_name = 'file_c.csv'",
                "  b.to_csv(c_name)",
                "  b.MyClass.save_file(d)",
            ],
        )
        ins, outs = find_dependencies(
            loader,
            self.file_path("a.py"),
            [ExprPattern.from_str(r'read_csv(r"\w+\.csv")')],
            [ExprPattern.from_str(r'`*`.to_csv(r"\w+\.csv")')],
        )
        self.assertEqual(ins, ["abc.csv", "file_b.csv", "qwe.csv"])
        self.assertEqual(outs, ["def.csv", "file_c.csv", "asd.csv"])

    def test_load_module_from_the_same_package(self):
        loader = Loader([self._dir.name])
        self.write_file(
            "a/b.py",
            [
                "from a import c",
                "",
            ],
        )
        self.write_file(
            "a/c.py",
            [
                'c = "abc"',
                "",
            ],
        )
        self.write_file(
            "a/__init__.py",
            [
                "from . import b",
                "",
            ],
        )
        self.write_file("a/__main__.py", ["print('executing')", ""])
        self.assertObjectEqual(
            build_module_from_filepath(loader, self.file_path("a")),
            Package(
                modules={
                    "__init__": Module(
                        ast=ast.Module(
                            body=[ast.ImportFrom(names=[ast.alias(name="b")], level=1)],
                            type_ignores=[],
                        ),
                        spec=self.spec("a"),
                        children={
                            "b": Module(
                                ast=ast.Module(
                                    body=[
                                        ast.ImportFrom(
                                            module="a",
                                            names=[ast.alias(name="c")],
                                            level=0,
                                        )
                                    ],
                                    type_ignores=[],
                                ),
                                spec=self.spec("a/b"),
                                children={
                                    "c": Module(
                                        ast.Module(
                                            body=[
                                                ast.Assign(
                                                    targets=[
                                                        ast.Name(
                                                            id="c", ctx=ast.Store()
                                                        )
                                                    ],
                                                    value=ast.Constant(value="abc"),
                                                )
                                            ],
                                            type_ignores=[],
                                        ),
                                        spec=self.spec("a/c"),
                                        loaded=True,
                                        children={
                                            "c": Node(
                                                ast.Constant(value="abc"),
                                                spec=self.spec("a/c"),
                                            )
                                        },
                                    )
                                },
                                loaded=True,
                            )
                        },
                    ),
                    "c": Module(
                        ast.Module(
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="c", ctx=ast.Store())],
                                    value=ast.Constant(value="abc"),
                                )
                            ],
                            type_ignores=[],
                        ),
                        spec=self.spec("a/c"),
                        loaded=True,
                        children={
                            "c": Node(ast.Constant(value="abc"), spec=self.spec("a/c"))
                        },
                    ),
                    "b": Module(
                        ast=ast.Module(
                            body=[
                                ast.ImportFrom(
                                    module="a", names=[ast.alias(name="c")], level=0
                                )
                            ],
                            type_ignores=[],
                        ),
                        spec=self.spec("a/b"),
                        loaded=True,
                        children={
                            "c": Module(
                                ast.Module(
                                    body=[
                                        ast.Assign(
                                            targets=[ast.Name(id="c", ctx=ast.Store())],
                                            value=ast.Constant(value="abc"),
                                        )
                                    ],
                                    type_ignores=[],
                                ),
                                spec=self.spec("a/c"),
                                loaded=True,
                                children={
                                    "c": Node(
                                        ast.Constant(value="abc"),
                                        spec=self.spec("a/c"),
                                    )
                                },
                            )
                        },
                    ),
                }
            ),
        )



import argparse

from deba.commands.decorators import subcommand
from deba.config import Config


def exec(conf: Config, args: argparse.Namespace):
    print(conf.data_dir)


@subcommand(exec=exec)
def add_subcommand(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(name="dataDir")
    return parser



import argparse

from .init import add_subcommand as add_init_command
from .data_dir import add_subcommand as add_data_dir_command
from .deps import add_subcommand as add_deps_command
from .stages import add_subcommand as add_stages_command
from .targets import add_subcommand as add_targets_command
from .python_path import add_subcommand as add_python_path_command


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        "deba", description="discover proper order of execution"
    )
    subparsers = parser.add_subparsers()
    add_init_command(subparsers)
    add_data_dir_command(subparsers)
    add_deps_command(subparsers)
    add_stages_command(subparsers)
    add_targets_command(subparsers)
    add_python_path_command(subparsers)
    return parser


__all__ = ["get_parser"]



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021 Wrangle Ltd

import typing
from unittest import TestCase

from attrs import define, field

from deba.serialize import yaml_load, yaml_dump


@define()
class Person(object):
    name: str = field(default=None)
    birth_date: str = field(default=None)
    height: int = field(default=None)
    scores: typing.List[int] = field(default=None)
    _gender: str = field(default=None)


@define()
class Website(object):
    url: str = field(default=None)


@define()
class Employment(object):
    title: str = field(default=None)


@define()
class Listing(object):
    user: Person = field(default=None)
    websites: typing.List[Website] = field(default=None)
    employments: typing.Dict[str, Employment] = field(default=None)


@define()
class Animal(object):
    omnivore: bool = field(default=None)
    vertebrate: bool = field(default=None)


class YAMLTestCase(TestCase):
    def test_loads_simple(self):
        obj1 = Person(
            name="John Doe", birth_date="10/02/2000", height=170, scores=[7, 8, 9]
        )
        obj1._gender = "male"
        yaml_str = yaml_dump(obj1)
        self.assertEqual(
            yaml_str,
            "\n".join(
                [
                    "birthDate: 10/02/2000",
                    "height: 170",
                    "name: John Doe",
                    "scores:",
                    "- 7",
                    "- 8",
                    "- 9",
                    "",
                ]
            ),
        )
        obj2 = yaml_load(yaml_str, Person)
        self.assertIsNone(obj2._gender)
        obj1._gender = None
        self.assertEqual(obj1, obj2)

    def test_loads_nested(self):
        self.maxDiff = None
        obj1 = Listing(
            user=Person(name="John Doe", birth_date="10/02/2000", height=170),
            websites=[
                Website(url="https://url1"),
                Website(url="https://url2"),
            ],
            employments={
                "ABC": Employment(title="engineer"),
                "DEF": Employment(title="head QA"),
            },
        )
        yaml_str = yaml_dump(obj1, indent=4)
        self.assertEqual(
            yaml_str,
            "\n".join(
                [
                    "employments:",
                    "    ABC:",
                    "        title: engineer",
                    "    DEF:",
                    "        title: head QA",
                    "user:",
                    "    birthDate: 10/02/2000",
                    "    height: 170",
                    "    name: John Doe",
                    "websites:",
                    "-   url: https://url1",
                    "-   url: https://url2",
                    "",
                ]
            ),
        )
        obj2 = yaml_load(yaml_str, Listing)
        self.assertEqual(obj1, obj2)

    def test_loads_ignore_empty(self):
        obj = yaml_load('{"omnivore": true}', Animal)
        self.assertEqual(obj, Animal(omnivore=True))



import argparse
import pathlib
import shutil

from deba.commands.decorators import subcommand
from deba.config import Config, Stage
from deba.deps.expr import ExprPatterns
from deba.serialize import yaml_dump


def is_line_found(file: pathlib.Path, expected_line: str) -> bool:
    if file.is_file():
        with open(file, "r") as f:
            for line in f:
                if line.strip() == expected_line:
                    return True
    return False


def ensure_line(file: pathlib.Path, line: str):
    if not is_line_found(file, line):
        with open(file, "a+") as f:
            f.write("\n%s\n" % line)
        print('added "%s" to %s' % (line, file.name))


def exec(conf: Config, args: argparse.Namespace):
    cwd = pathlib.Path.cwd()
    deba_file = cwd / "deba.yaml"
    if not deba_file.is_file():
        # prompt for stages
        stages = []
        if args.stages is None:
            while True:
                cont = input("add a stage? (Y/n)")
                if cont != "" and cont.strip().lower() != "y":
                    break
                stages.append(Stage(name=input("  stage name:")))
            if len(stages) == 0:
                raise ValueError("must add at least 1 stage")
        else:
            stages = [Stage(name=name.strip()) for name in args.stages]

        # prompt for targets
        targets = []
        if args.targets is None:
            while True:
                cont = input("add a target? (Y/n)")
                if cont != "" and cont.strip().lower() != "y":
                    break
                targets.append(input("  target: ").strip())
            if len(targets) == 0:
                raise ValueError("must add at least 1 target")
        else:
            targets = args.targets

        # prompt for patterns
        prerequisite_patterns = []
        if args.prerequisite_patterns is None:
            while True:
                cont = input("add an prerequisite pattern? (Y/n)")
                if cont != "" and cont.strip().lower() != "y":
                    break
                prerequisite_patterns.append(input("  prerequisite pattern: ").strip())
            if len(prerequisite_patterns) == 0:
                raise ValueError("must add at least 1 prerequisite pattern")
        else:
            prerequisite_patterns = args.prerequisite_patterns
        target_patterns = []
        if args.target_patterns is None:
            while True:
                cont = input("add an target pattern? (Y/n)")
                if cont != "" and cont.strip().lower() != "y":
                    break
                target_patterns.append(input("  target pattern: ").strip())
            if len(target_patterns) == 0:
                raise ValueError("must add at least 1 target pattern")
        else:
            target_patterns = args.target_patterns

        # write deba config
        conf = Config(
            stages=stages,
            targets=targets,
            patterns=ExprPatterns(
                prerequisites=prerequisite_patterns, targets=target_patterns
            ),
        )
        with open(deba_file, "w") as f:
            f.write(yaml_dump(conf))
        print("wrote deba config to %s" % deba_file.name)
    else:
        print("deba config found, skipping config initialization")

    # write make config
    mk_file = cwd / "deba.mk"
    shutil.copyfile(pathlib.Path(__file__).parent / "Makefile", mk_file)
    print("wrote Make config to %s" % mk_file.name)
    ensure_line(cwd / "Makefile", "include deba.mk")

    # write .gitignore
    ensure_line(cwd / ".gitignore", ".deba")


@subcommand(exec=exec, open_config=False)
def add_subcommand(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        name="init", description="initialize deba config in the current folder"
    )
    parser.add_argument(
        "--stages",
        type=str,
        nargs="*",
        help="names of execution stages. Each stage is a plain folder that houses scripts that have similar execution order. Execution order among stages follows their order in the config file.",
    )
    parser.add_argument(
        "--targets",
        type=str,
        nargs="*",
        help="target files. Each time your run `make deba`, these files will be updated if any of their dependencies have been updated since.",
    )
    parser.add_argument(
        "--prerequisite-patterns",
        type=str,
        nargs="*",
        help="prerequisite patterns that Dirk uses to find prerequisites for each script",
    )
    parser.add_argument(
        "--target-patterns",
        type=str,
        nargs="*",
        help="target patterns that Dirk uses to find targets for each script",
    )
    return parser



import argparse
import io
import json
import os
import typing

from deba.commands.decorators import subcommand
from deba.config import Config, Stage

from deba.deps.module import Loader
from deba.deps.find import find_dependencies


class InvalidDependencyError(Exception):
    pass


def validate_prerequisites(
    conf: Config, stage: Stage, ins: typing.List[str], rel_script_path: str
):
    ins_set = set()
    for filename in ins:
        if filename in ins_set:
            print(
                "WARNING: prerequisite %s of script %s found more than once"
                % (json.dumps(filename), rel_script_path)
            )
        ins_set.add(filename)
        if conf.is_data_from_latter_stages(stage.name, filename):
            raise InvalidDependencyError(
                "prerequisite %s of script %s comes from a later stage"
                % (json.dumps(filename), rel_script_path)
            )


def validate_targets(
    conf: Config, stage: Stage, outs: typing.List[str], rel_script_path: str
):
    outs_set = set()
    for filename in outs:
        if filename in outs_set:
            print(
                "WARNING: target %s of script %s found more than once"
                % (json.dumps(filename), rel_script_path)
            )
        outs_set.add(filename)
        if not filename.startswith(stage.name + "/"):
            raise InvalidDependencyError(
                "target %s of script %s must start with %s"
                % (
                    json.dumps(filename),
                    rel_script_path,
                    json.dumps(stage.name + "/"),
                )
            )


def write_deps(
    conf: Config,
    stage: Stage,
    loader: Loader,
    deps_file: io.TextIOWrapper,
    script_name: str,
    script_path: str,
):
    prerequisites, targets = find_dependencies(
        loader,
        script_path,
        conf.patterns.prerequisites or [],
        conf.patterns.targets or [],
    )

    if stage.ignored_targets is not None:
        targets = [s for s in targets if s not in stage.ignored_targets]

    rel_script_path = os.path.join(stage.name, script_name)
    validate_prerequisites(conf, stage, prerequisites, rel_script_path)
    validate_targets(conf, stage, targets, rel_script_path)

    if conf.overrides is not None:
        for idx, exec_rule in enumerate(conf.overrides):
            if exec_rule.target_set == set(targets):
                print(
                    "    override #%d matches targets, skipping script %s"
                    % (idx, script_name)
                )
                return

    # write rule for this script
    targets = " ".join(["$(DEBA_DATA_DIR)/%s" % name for name in targets])
    deps_file.write(
        "%s &: %s %s | $(DEBA_DATA_DIR)/%s\n\t$(call deba_execute,%s)\n\n"
        % (
            targets,
            "$(DEBA_MD5_DIR)/%s.md5" % (rel_script_path),
            " ".join(
                ["$(DEBA_DATA_DIR)/%s" % name for name in prerequisites]
                + (
                    [str(p) for p in stage.common_prerequisites]
                    if stage.common_prerequisites is not None
                    else []
                )
            ),
            stage.name,
            rel_script_path,
        )
    )


def exec(conf: Config, args: argparse.Namespace):
    if args.stage != "":
        loader = Loader(conf.script_search_paths)
        stage = conf.get_stage(args.stage)
        if stage is None:
            raise ValueError(
                "stage %s not found, available stages are: %s",
                json.dumps(args.stage),
                json.dumps([st.name for st in conf.stages]),
            )
        os.makedirs(conf.deps_dir, exist_ok=True)
        with open(stage.deps_filepath, "w") as f:
            # write rule for data dir
            f.write(
                "$(DEBA_DATA_DIR)/%s: ; @-mkdir -p $@ 2>/dev/null\n\n" % (stage.name)
            )

            for script_name, script_path in stage.scripts():
                write_deps(conf, stage, loader, f, script_name, script_path)
    else:
        os.makedirs(conf.deba_dir, exist_ok=True)
        with open(conf.main_deps_filepath, "w") as f:
            if conf.overrides is not None:
                for rule in conf.overrides:
                    f.write(
                        "%s &: %s\n\t%s\n\n"
                        % (
                            rule.target_str,
                            " ".join(
                                "$(DEBA_DATA_DIR)/%s" % d for d in rule.prerequisites
                            ),
                            rule.recipe,
                        )
                    )


@subcommand(exec=exec)
def add_subcommand(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(name="deps", description="write make rules")
    parser.add_argument(
        "--stage",
        type=str,
        default="",
        help="if specified, analyze and write make rules for scripts in this stage. Otherwise, write overriden make rules.",
    )
    return parser



import ast
import typing
import os
from importlib.machinery import ModuleSpec, PathFinder

import zope.interface
from zope.interface.common.collections import IMapping
from attrs import define, field


@zope.interface.implementer(IMapping)
@define
class Node(object):
    ast: ast.AST
    spec: ModuleSpec
    children: typing.Union[typing.Dict[str, "Node"], None] = field(default=None)

    def __getitem__(self, key):
        if self.children is not None:
            return self.children[key]
        raise KeyError("key %s not found" % key)

    def __contains__(self, key):
        if self.children is not None:
            return key in self.children
        return False

    def get(self, key, default=None):
        if self.children is not None:
            return self.children.get(key, default)
        return default

    def keys(self):
        if self.children is not None:
            return self.children.keys()
        return []

    def items(self):
        if self.children is not None:
            return self.children.items()
        return []

    def values(self):
        if self.children is not None:
            return self.children.values()
        return []


@zope.interface.implementer(IMapping)
@define
class Module(object):
    ast: ast.Module
    spec: ModuleSpec
    children: typing.Dict[str, Node] = field(factory=dict)
    loaded: bool = field(default=False)

    def __getitem__(self, key):
        return self.children[key]

    def __contains__(self, key):
        return key in self.children

    def get(self, key, default=None):
        return self.children.get(key, default)

    def keys(self):
        return self.children.keys()

    def items(self):
        return self.children.items()

    def values(self):
        return self.children.values()


@zope.interface.implementer(IMapping)
@define
class Package(object):
    modules: typing.Dict[str, typing.Union[Module, "Package"]] = field(factory=dict)

    def __getitem__(self, key):
        if key in self.modules["__init__"]:
            return self.modules["__init__"][key]
        return self.modules[key]

    def __contains__(self, key):
        if key in self.modules["__init__"]:
            return True
        return key in self.modules

    def get(self, key, default=None):
        if key in self.modules["__init__"]:
            return self.modules["__init__"][key]
        return self.modules.get(key, default)

    def keys(self):
        for k in self.modules["__init__"].keys():
            yield k
        for k in self.modules.keys():
            if k != "__init__":
                yield k

    def items(self):
        for k, v in self.modules["__init__"].items():
            yield k, v
        for k, v in self.modules.items():
            if k != "__init__":
                yield k, v

    def values(self):
        for k, v in self.modules["__init__"].items():
            yield v
        for k, v in self.modules.items():
            if k != "__init__":
                yield v

    @classmethod
    def from_spec(
        cls, loader: "Loader", spec: ModuleSpec, ast: ast.Module
    ) -> "Package":
        package = Package()
        package.modules["__init__"] = Module(ast, spec)
        package.collect_submodules(loader, spec)
        return package

    def collect_submodules(self, loader: "Loader", spec: ModuleSpec):
        for filename in os.listdir(spec.submodule_search_locations[0]):
            if filename in ["__init__.py", "__main__.py"]:
                continue
            if filename.endswith(".py"):
                name = trim_suffix(filename, ".py")
                self.modules[name] = loader.find_module(
                    name, spec.submodule_search_locations
                )
            elif os.path.isdir(
                os.path.join(spec.submodule_search_locations[0], filename)
            ) and os.path.exists(
                os.path.join(
                    spec.submodule_search_locations[0], filename, "__init__.py"
                )
            ):
                self.modules[filename] = loader.find_module(
                    filename, spec.submodule_search_locations
                )


@define
class Stack(object):
    layers: typing.List[typing.Dict[str, object]] = field(factory=lambda: [dict()])

    def push(self) -> "Stack":
        return Stack(self.layers + [dict()])

    def pop(self) -> "Stack":
        return Stack(self.layers[:-1])

    def store(self, key: str, node):
        self.layers[-1][key] = node

    def current_scope(self) -> dict:
        return self.layers[-1].copy()

    def remove(self, key: str):
        self.layers[-1].pop(key)

    def current_keys(self) -> typing.List[str]:
        return list(self.layers[-1].keys())

    def get_value(self, dotted_path: str) -> typing.Union[object, None]:
        parts = dotted_path.split(".")
        for scope in reversed(self.layers):
            if parts[0] not in scope:
                continue
            for name in parts:
                if scope is None:
                    return None
                node = scope.get(name, None)
                if node is None:
                    return None
                scope = node
            return node

    def dereference(self, t: ast.AST) -> typing.Union[Node, None]:
        if isinstance(t, ast.Name):
            return self.get_value(t.id)
        elif isinstance(t, ast.Attribute):
            stack: typing.List[ast.Attribute] = []
            expr = t
            while isinstance(expr, ast.Attribute):
                stack.append(expr)
                expr = expr.value
            if not isinstance(expr, ast.Name):
                return None
            node = self.get_value(expr.id)
            if node is not None:
                while len(stack) > 0:
                    attr = stack.pop()
                    node = Stack([node]).get_value(attr.attr)
                    if node is None:
                        break
            return node


def trim_suffix(s: str, suffix: str) -> str:
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


class ParseError(Exception):
    pass


@define
class Loader(object):
    paths: typing.List[str]
    module_asts: typing.Dict[str, ast.Module] = field(factory=dict)
    module_nodes: typing.Dict = field(factory=dict)

    def find_spec(
        self,
        module_name: str,
        parent_module_paths: typing.Union[typing.List[str], None] = None,
    ) -> typing.Union[ModuleSpec, None]:
        parts = module_name.split(".")
        paths = list(
            set(
                self.paths
                if parent_module_paths is None
                else parent_module_paths + self.paths
            )
        )
        for name in parts:
            spec = PathFinder.find_spec(name, paths)
            if spec is None:
                return None
            paths = getattr(spec, "submodule_search_locations", [])
        return spec

    def parse_ast(self, origin: str) -> ast.Module:
        if origin in self.module_asts:
            return self.module_asts[origin]
        with open(origin, "r") as f:
            try:
                root = ast.parse(f.read(), os.path.split(origin)[-1])
            except Exception as e:
                raise ParseError("error parsing %s" % origin, e)
            self.module_asts[origin] = root
            return root

    def find_module(
        self,
        module_name: str,
        parent_module_paths: typing.Union[typing.List[str], None] = None,
    ) -> typing.Union[object, None]:
        spec = self.find_spec(module_name, parent_module_paths)
        if spec is None or spec.origin is None or spec.origin.endswith(".pyc"):
            return None
        if spec.origin in self.module_nodes:
            return self.module_nodes[spec.origin]
        mod = self.parse_ast(spec.origin)
        if os.path.split(spec.origin)[-1] == "__init__.py":
            node = Package.from_spec(self, spec, mod)
            self.module_nodes[spec.origin] = node
            self.populate_module_scope(node.modules["__init__"])
        else:
            node = Module(mod, spec)
            self.module_nodes[spec.origin] = node
            self.populate_module_scope(node)
        return node

    def populate_module_scope(self, mod: Module):
        if mod.loaded:
            return
        mod.loaded = True
        stack = Stack()
        for stmt in mod.ast.body:
            self.populate_scope(mod.spec, stack, mod.ast, stmt)
        mod.children = stack.current_scope()

    def populate_scope(
        self,
        spec: ModuleSpec,
        stack: Stack,
        parent_node: ast.AST,
        stmt: ast.stmt,
    ):
        if isinstance(stmt, ast.Import):
            for alias in stmt.names:
                node = self.find_module(alias.name, spec.submodule_search_locations)
                if node is not None:
                    stack.store(
                        alias.name if alias.asname is None else alias.asname, node
                    )
        elif isinstance(stmt, ast.ImportFrom):
            if stmt.module is None:
                level = stmt.level
                package_path = spec.origin
                while level > 0:
                    package_path = os.path.dirname(package_path)
                    level -= 1
                parent, package_name = os.path.split(package_path)
                node = self.find_module(package_name, [parent])
            else:
                node = self.find_module(stmt.module, spec.submodule_search_locations)
            if node is not None:
                for alias in stmt.names:
                    value = Stack([node]).get_value(alias.name)
                    if value is not None:
                        stack.store(
                            alias.name if alias.asname is None else alias.asname,
                            value,
                        )
        elif isinstance(stmt, ast.FunctionDef) or isinstance(
            stmt, ast.AsyncFunctionDef
        ):
            # ignore instance methods
            if isinstance(parent_node, ast.ClassDef):
                instmed = True
                for name in stmt.decorator_list:
                    if name.id in ["classmethod", "staticmethod"]:
                        instmed = False
                        break
                if instmed:
                    return
            stack.store(stmt.name, Node(stmt, spec))
        elif isinstance(stmt, ast.ClassDef):
            if hasattr(stmt, "body"):
                substack = stack.push()
                for cd_stmt in stmt.body:
                    self.populate_scope(spec, substack, stmt, cd_stmt)
                children = substack.current_scope()
            else:
                children = None
            value = Node(stmt, spec, children)
            stack.store(stmt.name, value)
        elif isinstance(stmt, ast.Assign):
            node = self.dereference(spec, stack, stmt.value)
            if node is None:
                return
            for tgt in stmt.targets:
                if isinstance(tgt, ast.Tuple):
                    if isinstance(node.ast, ast.Tuple):
                        for idx, el in enumerate(tgt.elts):
                            if isinstance(el, ast.Name):
                                stack.store(el.id, Node(node.ast.elts[idx], spec))
                elif isinstance(tgt, ast.Name):
                    stack.store(tgt.id, Node(node.ast, spec))

    def dereference(
        self, spec: ModuleSpec, stack: Stack, node: ast.AST
    ) -> typing.Union[Node, None]:
        if isinstance(node, ast.Constant):
            return Node(node, spec)
        elif isinstance(node, ast.Tuple):
            for idx, el in enumerate(node.elts):
                child = self.dereference(spec, stack, el)
                if child is None:
                    return None
                node.elts[idx] = child.ast
            return Node(node, spec)
        else:
            return stack.dereference(node)



import argparse
import os

from deba.commands.decorators import subcommand
from deba.config import Config


def exec(conf: Config, args: argparse.Namespace):
    if conf.targets is None:
        return
    for target in conf.targets:
        print(os.path.join(conf.data_dir, target))


@subcommand(exec=exec)
def add_subcommand(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(name="targets", help="print targets")
    return parser



import unittest
from unittest.mock import patch, call
import argparse

from deba.commands.stages import add_subcommand
from deba.config import Config, Stage


class StagesCommandTestCase(unittest.TestCase):
    @patch("builtins.print")
    def test_run(self, mock_print):
        parser = argparse.ArgumentParser("deba")
        subparsers = parser.add_subparsers()
        add_subcommand(subparsers)
        conf = Config(stages=[Stage(name="clean"), Stage(name="fuse")])

        args = parser.parse_args(["stages"])
        args.exec(conf, args)

        mock_print.assert_has_calls([call("clean"), call("fuse")])



from importlib.machinery import ModuleSpec, SourceFileLoader
from unittest import TestCase
import re
import typing
import ast

from attrs import define
from deba.deps.expr import ExprPattern, ExprTemplateParseError
from deba.deps.module import Node, Stack
from deba.test_utils import ASTMixin


class ExprTemplateTestCase(ASTMixin, TestCase):
    def test_from_str(self):
        with self.assertRaises(ExprTemplateParseError) as cm:
            ExprPattern.from_str("`abc*")
        self.assertEqual(
            cm.exception.args, ("backtick pattern not closed at line 1, offset 1",)
        )

        with self.assertRaises(ExprTemplateParseError) as cm:
            ExprPattern.from_str("abc\n123")
        self.assertEqual(cm.exception.args, ("expect exactly 1 expression, found 2",))

        for s in ["abc('d', 'e')", "abc()", "abc(de=123, qw='asd')"]:
            with self.assertRaises(ExprTemplateParseError) as cm:
                ExprPattern.from_str(s)
            self.assertEqual(
                cm.exception.args,
                (
                    "function call must have exactly one argument or one keyword argument",
                ),
            )

        with self.assertRaises(ExprTemplateParseError) as cm:
            ExprPattern.from_str(r"r'*'")
        self.assertTrue("expression must be a function call" in cm.exception.args[0])

        with self.assertRaises(ExprTemplateParseError) as cm:
            ExprPattern.from_str(r"to_csv(r'*')")
        self.assertEqual(
            cm.exception.args,
            ("invalid regular expression r'*': nothing to repeat at position 0",),
        )

        with self.assertRaises(SyntaxError):
            ExprPattern.from_str(r"@@")

        @define
        class Case:
            s: str
            expr: ast.AST
            file_pat: object
            patterns: typing.List[str]

        for case in [
            Case(
                r"load_csv('.+\\.csv')",
                ast.Call(
                    ast.Name("load_csv", ctx=ast.Load()),
                    args=[ast.Constant(r".+\.csv")],
                    keywords=[],
                ),
                re.compile(r"^.+\.csv$"),
                [],
            ),
            Case(
                r'`*`.to_csv("\\w+\\.csv")',
                ast.Call(
                    ast.Attribute(
                        ast.Name(
                            "deba_backtick_pat_000",
                            ctx=ast.Load(),
                        ),
                        attr="to_csv",
                        ctx=ast.Load(),
                    ),
                    args=[ast.Constant(r"\w+\.csv")],
                    keywords=[],
                ),
                re.compile(r"^\w+\.csv$"),
                ["*"],
            ),
        ]:
            et = ExprPattern.from_str(case.s)
            self.assertASTEqual(et.node, case.expr)
            self.assertEqual(et.file_pat, case.file_pat, repr(case))
            self.assertEqual(et.patterns, case.patterns, repr(case))

    def test_match_node(self):
        @define
        class Case:
            template: str
            source: str
            file: typing.Union[None, str]

        for idx, case in enumerate(
            [
                Case(r"read_csv(r'.+\.csv')", "read_csv('abc.csv')", "abc.csv"),
                Case(r"read_csv(r'.+\.csv')", "read_csv('abc.csvl')", None),
                Case(r"read_csv(r'.+\.csv')", "read_csv('abc.det')", None),
                Case(r"`*`.to_csv(r'.+\.csv')", "df.to_csv('abc.csv')", "abc.csv"),
                Case(
                    r"`*`.to_csv(r'.+\.csv')",
                    "df.to_csv('abc.csv', index=False)",
                    "abc.csv",
                ),
                Case(r"`*`.to_csv(r'.+\.csv')", "to_csv('abc.csv')", None),
                Case(r"`*`.to_csv(r'.+\.csv')", "df.load('abc.csv')", None),
                Case(r"`*`.to_csv(r'.+\.csv')", "df.to_csv('abc.cst')", None),
                Case(r"`save_*`(r'.+\.csv')", "save_user('abc.csv')", "abc.csv"),
                Case(r"`save_*`(r'.+\.csv')", "load_user('abc.csv')", None),
                Case(r"`save_*`(r'.+\.csv')", "save_user('abc.dsv')", None),
                Case(
                    r"`do_*`(`does_*`(kw=r'.+\.pdf'))",
                    "do_abc(123, does_xyz('abc', kw='qwe.pdf'))",
                    "qwe.pdf",
                ),
            ]
        ):
            et = ExprPattern.from_str(case.template)
            node = ast.parse(case.source).body[0].value
            self.assertEqual(
                et.match_node(Stack(), node),
                case.file,
                "(case %d) %s" % (idx, repr(case)),
            )

    def test_match_node_with_scopes(self):
        spec = ModuleSpec("a", loader=SourceFileLoader("a", "a"))
        scopes = Stack(
            [
                {
                    "a": Node(ast.Constant(value="file_a.csv"), spec),
                    "MyClass": Node(
                        ast.ClassDef(
                            name="MyClass",
                            bases=[],
                            keywords=[],
                            body=[
                                ast.Assign(
                                    targets=[ast.Name(id="b", ctx=ast.Store())],
                                    value=ast.Constant(value="file_b.csv"),
                                ),
                                ast.Assign(
                                    targets=[ast.Name(id="c", ctx=ast.Store())],
                                    value=ast.Constant(value="file_c.pdf"),
                                ),
                            ],
                        ),
                        spec,
                        children={
                            "b": Node(
                                ast.Constant(value="file_b.csv"),
                                spec,
                            ),
                            "c": Node(
                                ast.Constant(value="file_c.pdf"),
                                spec,
                            ),
                        },
                    ),
                }
            ]
        )

        @define
        class Case:
            template: str
            source: str
            file: typing.Union[None, str]

        for idx, case in enumerate(
            [
                Case(r"read_csv(r'.+\.csv')", "read_csv(a)", "file_a.csv"),
                Case(r"read_csv(r'.+\.csv')", "read_csv(d)", None),
                Case(r"read_csv(r'.+\.csv')", "read_csv(MyClass.b)", "file_b.csv"),
                Case(r"read_csv(r'.+\.csv')", "read_csv(MyClass.c)", None),
            ]
        ):
            et = ExprPattern.from_str(case.template)
            node = ast.parse(case.source).body[0].value
            self.assertEqual(
                et.match_node(scopes, node),
                case.file,
                "(case %d) %s" % (idx, repr(case)),
            )



import argparse
import unittest
import unittest.mock

from deba.commands.decorators import subcommand
from deba.config import Config, Stage


def spy():
    def inner(func):
        return unittest.mock.Mock(wraps=func)

    return inner


class WithConfigTestCase(unittest.TestCase):
    def test_decorator(self):
        my_conf = Config(stages=[Stage(name="clean")], targets=["clean/my_data.csv"])

        @spy()
        def exec(conf: Config, args: argparse.Namespace):
            self.assertEqual(conf, my_conf)
            self.assertEqual(args.my_arg, 3)
            self.assertEqual(args.my_sub_flag, "abc")

        @subcommand(exec=exec)
        def add_my_subcommand(
            subparsers: argparse._SubParsersAction,
        ) -> argparse.ArgumentParser:
            parser = subparsers.add_parser(name="my-sub-command")
            parser.add_argument("--my-sub-flag", type=str, default="def")
            return parser

        parser = argparse.ArgumentParser("my-program")
        parser.set_defaults(my_arg=3)
        subparsers = parser.add_subparsers()
        add_my_subcommand(subparsers)

        args = parser.parse_args(["my-sub-command", "--my-sub-flag", "abc"])
        args.exec(my_conf, args)
        exec.assert_called_once_with(my_conf, args)



import unittest
from unittest.mock import patch
import argparse

from deba.commands.data_dir import add_subcommand
from deba.config import Config, Stage


class DataDirCommandTestCase(unittest.TestCase):
    @patch("builtins.print")
    def test_run(self, mock_print):
        parser = argparse.ArgumentParser("deba")
        subparsers = parser.add_subparsers()
        add_subcommand(subparsers)
        conf = Config(stages=[Stage(name="clean")])

        args = parser.parse_args(["dataDir"])
        args.exec(conf, args)

        mock_print.assert_called_with("data")



import argparse
import typing

from deba.config import Config, get_config

EXEC_FUNC = typing.Callable[[Config, argparse.Namespace], None]
ADD_COMMAND_FUNC = typing.Callable[
    [argparse._SubParsersAction], argparse.ArgumentParser
]


def subcommand(exec: EXEC_FUNC, open_config=True):
    def inner(func: ADD_COMMAND_FUNC):
        def inner_still(
            subparsers: argparse._SubParsersAction,
        ):
            parser = func(subparsers)

            def _exec(conf: Config, args: argparse.Namespace):
                if open_config and conf is None:
                    conf = get_config()
                exec(conf, args)

            parser.set_defaults(exec=_exec)

        return inner_still

    return inner



import unittest
import argparse

from deba.commands.deps import add_subcommand
from deba.config import Config, ExecutionRule, Stage
from deba.deps.expr import ExprPatterns
from deba.test_utils import TempDirMixin


class DepsCommandTestCase(TempDirMixin, unittest.TestCase):
    def test_run(self):
        parser = argparse.ArgumentParser("deba")
        subparsers = parser.add_subparsers()
        add_subcommand(subparsers)
        conf = Config(
            stages=[
                Stage(name="clean", ignored_scripts=["d.py"]),
                Stage(name="fuse", ignored_targets=["duplicates.csv"]),
            ],
            targets=["fuse/data.csv"],
            patterns=ExprPatterns(
                prerequisites=[r'read_csv(".+\\.csv")'],
                targets=[r'`*`.to_csv(".+\\.csv")'],
            ),
            overrides=[
                ExecutionRule(
                    target="clean/b_output.csv",
                    prerequisites=["raw/my_b_input.csv"],
                    recipe="my_command",
                ),
                ExecutionRule(
                    target="clean/c.csv",
                    prerequisites=["raw/c.csv"],
                    recipe="my_other_command",
                ),
            ],
            root_dir=self._dir.name,
        )

        self.write_file(
            "clean/a.py",
            [
                'if __name__ == "__main__":',
                '  df = read_csv("raw/a_input.csv")',
                '  df.to_csv("clean/a_output.csv")',
            ],
        )
        self.write_file(
            "clean/b.py",
            [
                'if __name__ == "__main__":',
                '  df = read_csv("raw/b_input.csv")',
                '  df.to_csv("clean/b_output.csv")',
            ],
        )
        self.write_file(
            "clean/d.py",
            [
                'if __name__ == "__main__":',
                '  df = read_csv("raw/d_input.csv")',
                '  df.to_csv("clean/d_output.csv")',
            ],
        )
        self.write_file(
            "fuse/a.py",
            [
                'if __name__ == "__main__":',
                '  df = read_csv("clean/b_output.csv")',
                '  df.to_csv("duplicates.csv")',
                '  df.to_csv("fuse/data.csv")',
            ],
        )
        self.write_file("fuse/b.pyc", [""])

        args = parser.parse_args(["deps", "--stage", "clean"])
        args.exec(conf, args)

        self.assertFileContent(
            ".deba/deps/clean.d",
            [
                "$(DEBA_DATA_DIR)/clean: ; @-mkdir -p $@ 2>/dev/null",
                "",
                "$(DEBA_DATA_DIR)/clean/a_output.csv &: $(DEBA_MD5_DIR)/clean/a.py.md5 $(DEBA_DATA_DIR)/raw/a_input.csv | $(DEBA_DATA_DIR)/clean",
                "\t$(call deba_execute,clean/a.py)",
                "",
                "",
            ],
        )

        args = parser.parse_args(["deps", "--stage", "fuse"])
        args.exec(conf, args)

        self.assertFileContent(
            ".deba/deps/fuse.d",
            [
                "$(DEBA_DATA_DIR)/fuse: ; @-mkdir -p $@ 2>/dev/null",
                "",
                "$(DEBA_DATA_DIR)/fuse/data.csv &: $(DEBA_MD5_DIR)/fuse/a.py.md5 $(DEBA_DATA_DIR)/clean/b_output.csv | $(DEBA_DATA_DIR)/fuse",
                "\t$(call deba_execute,fuse/a.py)",
                "",
                "",
            ],
        )

        args = parser.parse_args(["deps"])
        args.exec(conf, args)

        self.assertFileContent(
            ".deba/main.d",
            [
                "$(DEBA_DATA_DIR)/clean/b_output.csv &: $(DEBA_DATA_DIR)/raw/my_b_input.csv",
                "\tmy_command",
                "",
                "$(DEBA_DATA_DIR)/clean/c.csv &: $(DEBA_DATA_DIR)/raw/c.csv",
                "\tmy_other_command",
                "",
                "",
            ],
        )



import unittest
from unittest.mock import patch, call
import argparse

from deba.commands.targets import add_subcommand
from deba.config import Config, Stage


class TargetsCommandTestCase(unittest.TestCase):
    @patch("builtins.print")
    def test_run(self, mock_print):
        parser = argparse.ArgumentParser("deba")
        subparsers = parser.add_subparsers()
        add_subcommand(subparsers)
        args = parser.parse_args(["targets"])

        args.exec(
            Config(
                stages=[Stage(name="clean"), Stage(name="fuse")],
            ),
            args,
        )
        mock_print.assert_not_called()

        args.exec(
            Config(
                stages=[Stage(name="clean"), Stage(name="fuse")],
                targets=["fuse/personnel.csv", "fuse/complaint.csv"],
            ),
            args,
        )
        mock_print.assert_has_calls(
            [call("data/fuse/personnel.csv"), call("data/fuse/complaint.csv")]
        )



import argparse
import os

from deba.commands.decorators import subcommand
from deba.config import Config


def exec(conf: Config, args: argparse.Namespace):
    print(os.pathsep.join(conf.script_search_paths))


@subcommand(exec=exec)
def add_subcommand(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(name="pythonPath", help="print pythonPath")
    return parser



import os
import typing
import pathlib

from attrs import define, validators

from deba.attrs_utils import field_transformer, doc
from deba.serialize import yaml_dump, yaml_load
from deba.deps.expr import ExprPatterns


@define(field_transformer=field_transformer(globals()), slots=False)
class Stage(object):
    """A stage is a group of scripts that have the same order of execution."""

    name: str = doc(
        "name of the stage",
        required=True,
        validator=validators.matches_re(r"^[a-zA-Z][a-zA-Z0-9-_]+$"),
    )
    ignored_scripts: typing.List[str] = doc(
        "list of scripts that will be ignored during dependency analysis"
    )
    common_prerequisites: typing.List[str] = doc(
        "list of common prerequisites of every scripts in this stage"
    )
    ignored_targets: typing.List[str] = doc(
        "list of targets that will be ignored (not written to Makefile)"
    )

    @property
    def deps_filepath(self) -> str:
        return os.path.join(self._conf.deps_dir, "%s.d" % self.name)

    @property
    def script_dir(self) -> str:
        return os.path.join(self._conf._root_dir, self.name)

    def scripts(self) -> typing.Iterator[str]:
        for filename in os.listdir(self.script_dir):
            if self.ignored_scripts is not None and filename in self.ignored_scripts:
                continue
            if filename.endswith(".py"):
                yield filename, os.path.join(self.script_dir, filename)


@define(field_transformer=field_transformer(globals()))
class ExecutionRule(object):
    """An execution rule defines how a target or a list of targets should be generated.

    An execution rule is roughly equivalent to a make rule like this:

        [target]: [prerequisites]
            [recipe]
    """

    target: typing.Union[str, typing.List[str]] = doc(
        "single target or list of targets that can be generated"
    )
    prerequisites: typing.List[str] = doc(
        "list of prerequisites that when newer than the targets, trigger the execution"
    )
    recipe: str = doc("the command to execute")

    @property
    def target_set(self) -> typing.Set[str]:
        if type(self.target) is str:
            return set([self.target])
        return set(self.target)

    @property
    def target_str(self) -> str:
        if type(self.target) is str:
            return "$(DEBA_DATA_DIR)/%s" % self.target
        return " ".join("$(DEBA_DATA_DIR)/%s" % s for s in self.target)


@define(field_transformer=field_transformer(globals()))
class Config(object):
    """Dirk configurations."""

    stages: typing.List[Stage] = doc(
        "list of execution stages. The order of this list is also the order of execution. What this mean concretely is that scripts from a stage can only take target from earlier stages or the same stage as prerequisite.",
        required=True,
    )

    root_dir: str = doc(
        "root directory from which to locate data and stage directories"
    )

    targets: typing.List[str] = doc(
        "explicit targets to generate when user run `make deba`"
    )
    patterns: ExprPatterns = doc("expression templates")
    overrides: typing.List[ExecutionRule] = doc(
        "list of make rule overrides. If a make rule with the same targets exists, replace it with the corresponding rule defined here."
    )
    python_path: typing.List[str] = doc(
        "additional search paths for module files. The directory that contains deba.yaml file will be prepended to this list. This list is then concatenated as PYTHONPATH env var during script execution."
    )
    data_dir: str = doc(
        "keep all generated data in this folder",
        default="data",
        converter=lambda s: s if type(s) is not str else s.strip("/"),
        required=True,
    )

    @property
    def script_search_paths(self) -> typing.List[str]:
        if self.python_path is None:
            return [self._root_dir]
        return [self._root_dir] + self.python_path

    def __attrs_post_init__(self):
        for stage in self.stages:
            stage._conf = self

    def get_stage(self, name: str) -> typing.Union[Stage, None]:
        for stage in self.stages:
            if stage.name == name:
                return stage

    def is_data_from_latter_stages(self, stage_name: str, file_name: str) -> bool:
        check = False
        file_stage = file_name.split("/")[0]
        for stage in self.stages:
            if stage.name == stage_name:
                check = True
            elif check and stage.name == file_stage:
                return True
        return False

    def save(self):
        with open(os.path.join(self._root_dir, "deba.yaml"), "w") as f:
            f.write(yaml_dump(self))

    @property
    def _root_dir(self) -> str:
        if self.root_dir is not None:
            return self.root_dir
        return os.getcwd()

    @property
    def deba_dir(self) -> str:
        return os.path.join(self._root_dir, ".deba")

    @property
    def deps_dir(self) -> str:
        return os.path.join(self.deba_dir, "deps")

    @property
    def main_deps_filepath(self) -> str:
        return os.path.join(self.deba_dir, "main.d")


_conf = None


def get_config() -> Config:
    global _conf
    if _conf is not None:
        return _conf
    try:
        with open("deba.yaml", "r") as f:
            _conf = yaml_load(f.read(), Config)
        return _conf
    except FileNotFoundError:
        raise FileNotFoundError(
            "deba config file not found: %s" % (pathlib.Path().cwd() / "deba.yaml")
        )
